CharacterInfoStorage = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceDebug-2.0")
CharacterInfoStorage.version = "0.1." .. string.sub("$Revision: 25921 $", 12, -3);
CharacterInfoStorage.date = string.sub("$Date: 2007-01-23 16:50:43 -0500 (Tue, 23 Jan 2007) $", 8, 17);
CharacterInfoStorage:RegisterDB("CharacterInfoStorageDB");
CharacterInfoStorage:RegisterDefaults('profile', {
	StoreInventory = true,
	StoreBags = true,
	StoreBank = true,
	StoreKeyring = false,
--	StoreMail = false,
--	StoreSkills = false,
--	StoreFaction = false,
--	StoreTradeskill = false,
--	StoreQuest = false,
});

BINDING_HEADER_CHARACTERINFO = "Character Info";
BINDING_NAME_TOGGLE_CHARACTERINFO = "Toggle Character Info Window";

CharacterInfoStorage.vars = {}

local function getIdFromLink(link)
  if not link then return end
  local _,_,id = strfind(link,"item:(%d+):")
  if id then return tonumber(id) end
 	local _,_,id = strfind(link,"enchant:(%d+)")
 	if id then return -tonumber(id) end
end

local function getItemStringFromLink(link)
  if not link then return end
  local _,_,itemString = strfind(link,"item:(%d+:%d+:%d+:%d+:%d+:%d+:%d+)")
	return itemString;
end

function CharacterInfoStorage:StoringInventory()
	return self.db.profile.StoreInventory;
end

function CharacterInfoStorage:StoringBags()
	return self.db.profile.StoreBags;
end

function CharacterInfoStorage:StoringBank()
	return self.db.profile.StoreBank;
end

function CharacterInfoStorage:StoringKeyring()
	return self.db.profile.StoreKeyring;
end

function CharacterInfoStorage:OnInitialize()
	self:RegisterChatCommand({"/CharacterInfoStorage", "/cis", "/ci"}, {
		desc = 'Stores information about all your characters',
		handler = CharacterInfoStorage,
		type = 'group',
		args = {
			menu = {
				name = 'CharacterInfo',
				desc = 'Show Character Info Menu',
				type = 'execute',
				func = function() CharacterInfoStorage:ToggleUI() end,
			},
			remove = {
				name = 'Remove',
				desc = 'Removes information about a character',
				type = 'text',
				usage = '<charactername>',
				get = false,
				set = function(v) CharacterInfoStorage:Remove(v) end,
			},
			update = {
				name = 'Update',
				desc = 'Force a complete update of this character',
				type = 'execute',
				func = function() CharacterInfoStorage:InitPlayer() end,
			},
			purge = {
				name = 'Purge',
				desc = 'Removes information that is not to be stored from all characters. Warning! Will use current setting to determing what to remove from all characters.',
				type = 'execute',
				func = function() CharacterInfoStorage:Purge() end,
			},
			inventory = {
				name = 'StoreInventory',
				desc = 'Store characters equipped items',
				type = 'toggle',
				get = function() return self.db.profile.StoreInventory end,
				set = function(val) self.db.profile.StoreInventory = val end,
			},
			bags = {
				name = 'StoreBags',
				desc = 'Store characters bag content',
				type = 'toggle',
				get = function() return self.db.profile.StoreBags end,
				set = function(val) self.db.profile.StoreBags = val end,
			},
			bank = {
				name = 'StoreBank',
				desc = 'Store characters bank content (including bank bags)',
				type = 'toggle',
				get = function() return self.db.profile.StoreBank end,
				set = function(val) self.db.profile.StoreBank = val end,
			},
			keyring = {
				name = 'StoreKeyring',
				desc = 'Store characters keyring content',
				type = 'toggle',
				get = function() return self.db.profile.StoreKeyring end,
				set = function(val) self.db.profile.StoreKeyring = val end,
			},
--			mail = {
--				name = 'StoreMail',
--				desc = 'Store characters mailbox content',
--				type = 'toggle',
--				get = function() return self.db.profile.StoreMail end,
--				set = function(val) self.db.profile.StoreMail = val end,
--			},
--			skills = {
--				name = 'StoreSkills',
--				desc = 'Store characters skills',
--				type = 'toggle',
--				get = function() return self.db.profile.StoreSkills end,
--				set = function(val) self.db.profile.StoreSkills = val end,
--			},
--			faction = {
--				name = 'StoreFaction',
--				desc = 'Store characters factions',
--				type = 'toggle',
--				get = function() return self.db.profile.StoreFaction end,
--				set = function(val) self.db.profile.StoreFaction = val end,
--			},
--			tradeskill = {
--				name = 'StoreTradeskill',
--				desc = 'Store characters known tradeskill recipes',
--				type = 'toggle',
--				get = function() return self.db.profile.StoreTradeskill end,
--				set = function(val) self.db.profile.StoreTradeskill = val end,
--			},
--			quest = {
--				name = 'StoreQuest',
--				desc = 'Store characters quests',
--				type = 'toggle',
--				get = function() return self.db.profile.StoreQuest end,
--				set = function(val) self.db.profile.StoreQuest = val end,
--			},
		},
	});
	self.vars.playerName = UnitName("player");
	self.vars.playerClass = UnitClass("player");
	self.vars.playerLevel = UnitLevel("player");
	self:UpgradeBankBags();
	self:UpgradeItemStrings();
end

function CharacterInfoStorage:OnEnable()
	self:RegisterEvent("AceEvent_FullyInitialized","OnFullyInitialized",true);
	self:RegisterEvent("PLAYER_LEVEL_UP","OnLevelUp");
	self:RegisterEvent("PLAYER_MONEY","OnMoneyUpdate");
	self:RegisterEvent("UNIT_INVENTORY_CHANGED","OnInventoryUpdate");
	self:RegisterEvent("BAG_UPDATE","OnBagUpdate");
	self:RegisterEvent("BANKFRAME_OPENED","OnBankShow");
	self:RegisterEvent("PLAYERBANKBAGSLOTS_CHANGED","OnBankBagUpdate");
	self:RegisterEvent("PLAYERBANKSLOTS_CHANGED","OnBankUpdate");
	--self:RegisterEvent("MAIL_SHOW","OnMailShow");
	--self:RegisterEvent("MAIL_INBOX_UPDATE","OnMailShow");
	--self:RegisterEvent("SKILL_LINES_CHANGED","OnSkillUpdate");
	--self:RegisterEvent("TRADE_SKILL_SHOW","OnTradeskillShow");
	--self:RegisterEvent("CRAFT_SHOW","OnCraftShow");
	--self:RegisterEvent("UNIT_QUEST_LOG_CHANGED","OnQuestUpdate");
	self:InitPlayer();
end

function CharacterInfoStorage:OnDisable()
end

function CharacterInfoStorage:OnFullyInitialized()
	self:OnInventoryUpdate("player");
	self:InitBagScan();
end

function CharacterInfoStorage:Remove(name)
	if name == UnitName("player") then
		self:Print("Can't remove current character.");
		return;
	end
	self.db.realm[name] = nil;
end

function CharacterInfoStorage:Purge()
	for name,char in pairs(self.db.realm) do
		if not self:StoringInventory() then
			char.inventory = {};
		end
		if not self:StoringBags() then
			char.bags = {};
		end
		if not self:StoringBank() then
			char.bank = {};
		end
		if not self:StoringKeyring() then
			char.keyring = {};
		end
	end
end

function CharacterInfoStorage:InitPlayer()
	if not self.db.realm[self.vars.playerName] then 
		self.db.realm[self.vars.playerName] = {};
	end
	self.db.realm[self.vars.playerName].class = UnitClass("player");
	self.db.realm[self.vars.playerName].level = UnitLevel("player");
	self.db.realm[self.vars.playerName].sex = UnitSex("player");
	self.db.realm[self.vars.playerName].race = UnitRace("player");
	self.db.realm[self.vars.playerName].money = GetMoney();
	
	if UnitPowerType("player") == 0 then
		self.db.realm[self.vars.playerName].mana = UnitManaMax("player");
	else
		self.db.realm[self.vars.playerName].mana = nil;
	end
	
	self.db.realm[self.vars.playerName].health = UnitHealthMax("player");
	
	local base, _, armor = UnitArmor("player")
	self.db.realm[self.vars.playerName].armor = tostring(base).."|"..tostring(armor);
	
	if not self.db.realm[self.vars.playerName].stats then 
		self.db.realm[self.vars.playerName].stats = {};
	end
	for j=1, 5 do
		local base,stat,posBuff,negBuff = UnitStat("player",j);
		self.db.realm[self.vars.playerName].stats[j]= tostring(base-posBuff-negBuff).."|"..tostring(stat);
	end
	
	if not self.db.realm[self.vars.playerName].resistance then 
		self.db.realm[self.vars.playerName].resistance = {};
	end
	for j=1, 5 do
		local base,stat,posBuff,negBuff = UnitResistance("player",j);
		self.db.realm[self.vars.playerName].resistance[j]= tostring(base).."|"..tostring(stat);
	end
	
	self.db.realm[self.vars.playerName].guildName, self.db.realm[self.vars.playerName].guildRankName = GetGuildInfo("player")

	if not self.db.realm[self.vars.playerName].inventory then 
		self.db.realm[self.vars.playerName].inventory = {};
	end
	self:OnInventoryUpdate("player");

	if not self.db.realm[self.vars.playerName].bags then 
		self.db.realm[self.vars.playerName].bags = {};
	end
	if not self.db.realm[self.vars.playerName].keyring then 
		self.db.realm[self.vars.playerName].keyring = {};
	end
	self:InitBagScan();

	if not self.db.realm[self.vars.playerName].bank then 
		self.db.realm[self.vars.playerName].bank = {};
	end

end

function CharacterInfoStorage:InitBagScan()
	for bag=0,4 do
		self:OnBagUpdate(bag);
	end
	self:OnBagUpdate(-2);
	self:UpdateInventoryBags();
end

function CharacterInfoStorage:OnLevelUp(level)
	self:InitPlayer();
	self.db.realm[self.vars.playerName].level = level;
end

function CharacterInfoStorage:OnMoneyUpdate()
	self.db.realm[self.vars.playerName].money = GetMoney();
end

function CharacterInfoStorage:OnInventoryUpdate(unitID)
	if unitID ~= "player" then return end
	self:Debug("OnInventoryUpdate");
	if self:StoringInventory() then
		for index = 1, 19 do
		  link = GetInventoryItemLink("player", index);
		  if link then
				local itemString = getItemStringFromLink(link);
	      local num = GetInventoryItemCount("player", index);
	      if num > 1 then
	      	self.db.realm[self.vars.playerName].inventory[index] = itemString.."|"..num;
	      else
	      	self.db.realm[self.vars.playerName].inventory[index] = itemString;
	      end
	    else
	    	self.db.realm[self.vars.playerName].inventory[index] = nil;
		  end
		end
		self:UpdateInventoryBags();
	end
end

function CharacterInfoStorage:OnBagUpdate(bag)
	self:Debug("OnBagUpdate",bag);
--	local bag = arg1;
	if ( not bag ) then return end
	local bags;
	local prefix;
	local numSlots = GetContainerNumSlots(bag);
	if bag == -1 or bag > 4 then  -- Bank
		if not self:StoringBank() then return end
		bags = self.db.realm[self.vars.playerName].bank;
		if bag == -1 then
			prefix = 0;
		else
			prefix = (bag - 4)*100;
		end
	elseif bag == -2 then  -- Keyring
		if not self:StoringKeyring() then return end
		bags = self.db.realm[self.vars.playerName].keyring;
		prefix = 0;
	else
		if not self:StoringBags() then return end
		bags = self.db.realm[self.vars.playerName].bags;
		prefix = bag*100;
	end
	for slot=1,numSlots do
		local link = GetContainerItemLink(bag, slot);
		if link then
			local itemString = getItemStringFromLink(link);
			local _, stack = GetContainerItemInfo(bag, slot);
			bags[prefix+slot] = itemString.."|"..stack;
		else
			bags[prefix+slot] = nil;
		end
	end
	self:UpdateInventoryBags();
end

function CharacterInfoStorage:UpdateInventoryBags()
	if self:StoringInventory() then
		for bag = 1, 4 do
			index = ContainerIDToInventoryID(bag)
		  link = GetInventoryItemLink("player", index);
		  if link then
				local itemString = getItemStringFromLink(link);
				local numSlots = GetContainerNumSlots(bag);
				self.db.realm[self.vars.playerName].inventory[index] = itemString.."|"..numSlots;
			else
				self.db.realm[self.vars.playerName].inventory[index] = nil;
			end
		end
	end
end

function CharacterInfoStorage:UpdateBankBags()
	if self:StoringBank() then
		for bag = 5, 11 do
			index = ContainerIDToInventoryID(bag)
		  link = GetInventoryItemLink("player", index);
		  if link then
				local itemString = getItemStringFromLink(link);
				local numSlots = GetContainerNumSlots(bag);
				self.db.realm[self.vars.playerName].bank[index] = itemString.."|"..numSlots;
			else
				self.db.realm[self.vars.playerName].bank[index] = nil;
			end
		end
	end
end
	
function CharacterInfoStorage:OnBankShow()
	self:Debug("OnBankShow");
	self:OnBagUpdate(-1);
	self:UpdateBankBags();
	for bag = 5, 11 do
		self:OnBagUpdate(bag);
	end
end

function CharacterInfoStorage:OnBankBagUpdate()
	self:Debug("OnBankBagUpdate");
end

function CharacterInfoStorage:OnBankUpdate()
	self:Debug("OnBankUpdate");
	self:OnBagUpdate(-1);
	self:UpdateBankBags();
end

--[[
function CharacterInfoStorage:OnMailShow()
	self:Debug("OnMailShow");
end

function CharacterInfoStorage:OnSkillUpdate()
	self:Debug("OnSkillUpdate");
end

function CharacterInfoStorage:OnTradeskillShow()
	self:Debug("OnTradeskillShow");
end

function CharacterInfoStorage:OnCraftShow()
	self:Debug("OnCraftShow");
end

function CharacterInfoStorage:OnQuestUpdate()
	self:Debug("OnQuestUpdate");
end
]]
----------------------------------------------------------------------
-- UI Load and Toggle
----------------------------------------------------------------------

function CharacterInfoStorage:LoadUI(quiet)
	if not IsAddOnLoaded("CharacterInfo") then
		loaded,reason = LoadAddOn("CharacterInfo");
		if not loaded then
			if not quiet then self:Print("Could not load the UI. Reason: ",reason) end
		end
		return loaded
	end
	return true
end

function CharacterInfoStorage:ToggleUI()
	if self:LoadUI() then
		CharacterInfo:Character_Toggle();
	end
end
			
----------------------------------------------------------------------
-- Public API functions
----------------------------------------------------------------------
function CharacterInfoStorage:GetCharacters()
	local t = {};
	for i in pairs(self.db.realm) do
		table.insert(t,i);
	end
	return t;
end

function CharacterInfoStorage:GetNumItems(charName,itemId)
	local num=0;
	if self.db.realm[charName] then
		num = self:GetNumItemsBank(charName,itemId);
		num = num + self:GetNumItemsBags(charName,itemId);
		num = num + self:GetNumItemsInventory(charName,itemId);
	end
	return num;
end

function CharacterInfoStorage:GetNumItemsBank(charName,itemId)
	local num=0;
	if self.db.realm[charName] then
		for key,val in pairs(self.db.realm[charName].bank) do
			local _,_,id,i = string.find(val,"(%d+):%d+:%d+:%d+:%d+:%d+:%d+|?(%d*)")
			id = tonumber(id);
			if id == itemId then
				--self:Print(val,id,i)
				if key > 63 and key < 100 then i=1 end -- Bags have size, not num.
				if i and i~="" then num=num+tonumber(i) else num=num+1 end
			end
		end
	end
	return num;
end

function CharacterInfoStorage:GetNumItemsBags(charName,itemId)
	local num=0;
	if self.db.realm[charName] then
		for _,val in pairs(self.db.realm[charName].bags) do
			local _,_,id,i = string.find(val,"(%d+):%d+:%d+:%d+:%d+:%d+:%d+|?(%d*)")
			id = tonumber(id);
			if id == itemId then
				--self:Print(val,id,i)
				if i and i~="" then num=num+tonumber(i) else num=num+1 end
			end
		end
	end
	return num;
end

function CharacterInfoStorage:GetNumItemsInventory(charName,itemId)
	local num=0;
	if self.db.realm[charName] then
		for _,val in pairs(self.db.realm[charName].inventory) do
			local _,_,id,i = string.find(val,"(%d+):%d+:%d+:%d+:%d+:%d+:%d+|?(%d*)")
			id = tonumber(id);
			if id == itemId then
				--self:Print(val,id,i)
				num=num+1;
			end
		end
	end
	return num;
end

function CharacterInfoStorage:GetSlotItem(charName,slot)
	if self.db.realm[charName] and self.db.realm[charName].inventory[slot] then
		local _,_,itemString,num = string.find(self.db.realm[charName].inventory[slot],"(%d+:%d+:%d+:%d+:%d+:%d+:%d+)|?(%d*)");
		return "item:"..itemString..":0";
	end
end

-- In WoW 2.0 bank bags have changed slotid from 64-69 to 68-74. (1 more slot)
function CharacterInfoStorage:UpgradeBankBags()
	for charName in pairs(self.db.realm) do
		if self.db.realm[charName].bank[64] then
			for i = 69,64,-1 do
				self.db.realm[charName].bank[i+4] = self.db.realm[charName].bank[i];
				self.db.realm[charName].bank[i] = nil;
			end
		end
	end
end

-- In WoW 2.0 the itemString have changed from 4 sets to 8 sets of numbers.
-- 1.* itemId:enchantId:suffixId:uniqueId
-- 2.* itemId:enchantId:?:?:?:?:suffixId:uniqueId
function CharacterInfoStorage:UpgradeItemStrings()
	for charName in pairs(self.db.realm) do
		for k,v in pairs(self.db.realm[charName].inventory) do
			self.db.realm[charName].inventory[k] = self:UpgradeItemString(v);
		end
		for k,v in pairs(self.db.realm[charName].bags) do
			self.db.realm[charName].bags[k] = self:UpgradeItemString(v);
		end
		for k,v in pairs(self.db.realm[charName].bank) do
			self.db.realm[charName].bank[k] = self:UpgradeItemString(v);
		end
	end
end

function CharacterInfoStorage:UpgradeItemString(itemString)
	if not string.find(itemString,"(%d+:%d+:%d+:%d+:%d+:%d+:%d+)(|?%d*)") then
		local _,_,a1,a2,n = string.find(itemString,"(%d+:%d+:)(%d+)(|?%d*)")
		itemString = a1.."0:0:0:0:"..a2..n;
	end
	return itemString
end